#! /usr/bin/env luajit


-----------------------------------------------------------------
--Original code by MC-CNN
-- Edited by Storm Westlie











-----------------------------------------------------------------



require 'torch'

io.stdout:setvbuf('no')
for i = 1,#arg do
   io.write(arg[i] .. ' ')
end
io.write('\n')
dataset = table.remove(arg, 1)
arch = table.remove(arg, 1)
assert(dataset=='modelfish')
assert(arch == 'fast')

cmd = torch.CmdLine()
cmd:option('-gpu', 1, 'gpu id')
cmd:option('-seed', 42, 'random seed')
cmd:option('-debug', false)
cmd:option('-d', 'kitti | mb')
cmd:option('-a', 'train_tr | train_all | test_te | test_all | submit | time | predict', 'train')
cmd:option('-net_fname', '')
cmd:option('-make_cache', false)
cmd:option('-use_cache', false)
cmd:option('-print_args', false)
cmd:option('-sm_terminate', '', 'terminate the stereo method after this step')
cmd:option('-sm_skip', '', 'which part of the stereo method to skip')
cmd:option('-tiny', false)
cmd:option('-subset', 1)
cmd:option('-left', '')
cmd:option('-right', '')
cmd:option('-disp_max', '')
cmd:option('-rect', 'imperfect')
cmd:option('-color', 'gray')
cmd:option('-epochs', 50, 'Number of epochs to train for')
cmd:option('-epoch_step', 15, '1/2 learningrate at every epoch_step after epoch 15')

if arch == 'fast' then
  if dataset == 'modelfish' then
      cmd:option('-at', 0)
      cmd:option('-m', 0.2, 'margin')
      cmd:option('-pow', 1)

      cmd:option('-l1', 4)
      cmd:option('-fm', 64)
      cmd:option('-ks', 3)
      cmd:option('-lr', 0.04)
      cmd:option('-bs', 128)
      cmd:option('-mom', 0.9)
      cmd:option('-true1', 0)
      cmd:option('-false1', 4)
      cmd:option('-false2', 10)

      cmd:option('-hflip', 0)
      cmd:option('-vflip', 0)
      cmd:option('-rotate', 0)
      cmd:option('-hscale', 1.0)
      cmd:option('-scale', 1)
      cmd:option('-trans', 0)
      cmd:option('-hshear', 0.0)
      cmd:option('-brightness', 1.0)
      cmd:option('-contrast', 1.0)
      cmd:option('-d_vtrans', 0)
      cmd:option('-d_rotate', 0)
      cmd:option('-d_hscale', 1)
      cmd:option('-d_hshear', 0)
      cmd:option('-d_brightness', 0.0)
      cmd:option('-d_contrast', 1)
   end
end

opt = cmd:parse(arg)

if opt.print_args then
   print((opt.ks - 1) * opt.l1 + 1, 'arch_patch_size')
   print(opt.l1, 'arch1_num_layers')
   print(opt.fm, 'arch1_num_feature_maps')
   print(opt.ks, 'arch1_kernel_size')
   print(opt.l2, 'arch2_num_layers')
   print(opt.nh2, 'arch2_num_units_2')
   print(opt.false1, 'dataset_neg_low')
   print(opt.false2, 'dataset_neg_high')
   print(opt.true1, 'dataset_pos_low')
   os.exit()
end

require 'cunn'
require 'cutorch'
require 'image'
require 'libadcensus'
require 'libcv'
require 'cudnn'
require 'optim'
cudnn.benchmark = true

include('Margin2.lua')
include('Normalize2.lua')
include('StereoJoin.lua')
include('StereoJoin1.lua')
include('SpatialConvolution1_fw.lua')
-- include('SpatialLogSoftMax.lua')

torch.manualSeed(opt.seed)
cutorch.manualSeed(opt.seed)
cutorch.setDevice(tonumber(opt.gpu))

testLogger = optim.Logger(paths.concat('net', 'test.log'))
testLogger:setNames{'% Average prediction accuracy over training set', '% Average prediction accuracy over validation set'}
testLogger.showPlot = false


cmd_str = dataset .. '_' .. arch
for i = 1,#arg do
   cmd_str = cmd_str .. '_' .. arg[i]
end

function isnan(n)
   return tostring(n) == tostring(0/0)
end

function fromfile(fname)
   local file = io.open(fname .. '.dim')
   local dim = {}
   for line in file:lines() do
      table.insert(dim, tonumber(line))
   end
   if #dim == 1 and dim[1] == 0 then
      return torch.Tensor()
   end

   local file = io.open(fname .. '.type')
   local type = file:read('*all')

   local x
   if type == 'float32' then
      x = torch.FloatTensor(torch.FloatStorage(fname))
   elseif type == 'int32' then
      x = torch.IntTensor(torch.IntStorage(fname))
   elseif type == 'int64' then
      x = torch.LongTensor(torch.LongStorage(fname))
   else
      print(fname, type)
      assert(false)
   end

   x = x:reshape(torch.LongStorage(dim))
   return x
end

function get_window_size(net)
   ws = 1
   for i = 1,#net.modules do
      local module = net:get(i)
      if torch.typename(module) == 'cudnn.SpatialConvolution' then
         ws = ws + module.kW - 1
      end
   end
   return ws
end

-- load training data
if dataset == 'modelfish' then
   height = 300
   width = 1330
   disp_max = 170
   n_te = 0
   n_input_plane = 1
   err_at = 3

   if opt.a == 'train_tr' or opt.a == 'train_all' or opt.a == 'test_te' or opt.a == 'test_all' or opt.a == 'submit' then
      if dataset == 'modelfish' then
         X0 = fromfile('data.modelfish2019/x0.bin')
         X1 = fromfile('data.modelfish2019/x1.bin')
         dispnoc = fromfile('data.modelfish2019/dispnoc.bin')
         metadata = fromfile('data.modelfish2019/metadata.bin')
         tr = fromfile('data.modelfish2019/tr.bin')
         te = fromfile('data.modelfish2019/te.bin')
         nnz_tr = fromfile('data.modelfish2019/nnz_tr.bin')
         nnz_te = fromfile('data.modelfish2019/nnz_te.bin')
      end
   end
end

function savePNG(fname, x, isvol)
   local pred
   local pred_jet = torch.Tensor(1, 3, x:size(3), x:size(4))

   if isvol == true then
      pred = torch.CudaTensor(1, 1, x:size(3), x:size(4))
      adcensus.spatial_argmin(x, pred)
   else
      pred = x:double():add(1)
   end
   adcensus.grey2jet(pred[{1,1}]:div(disp_max):double(), pred_jet)
   image.savePNG(fname, pred_jet[1])
end

function saveOutlier(fname, x0, outlier)
   local img = torch.Tensor(1,3,height,img_width)
   img[{1,1}]:copy(x0)
   img[{1,2}]:copy(x0)
   img[{1,3}]:copy(x0)
   for i=1,height do
      for j=1,img_width do
         if outlier[{1,1,i,j}] == 1 then
            img[{1,1,i,j}] = 0
            img[{1,2,i,j}] = 1
            img[{1,3,i,j}] = 0
         elseif outlier[{1,1,i,j}] == 2 then
            img[{1,1,i,j}] = 1
            img[{1,2,i,j}] = 0
            img[{1,3,i,j}] = 0
         end
      end
   end
   image.savePNG(fname, img[1])
end

function gaussian(sigma)
   local kr = math.ceil(sigma * 3)
   local ks = kr * 2 + 1
   local k = torch.Tensor(ks, ks)
   for i = 1, ks do
      for j = 1, ks do
         local y = (i - 1) - kr
         local x = (j - 1) - kr
         k[{i,j}] = math.exp(-(x * x + y * y) / (2 * sigma * sigma))
      end
   end
   return k
end

function print_net(net)
   local s
   local t = torch.typename(net) 
   if t == 'cudnn.SpatialConvolution' then
      print(('conv(in=%d, out=%d, k=%d)'):format(net.nInputPlane, net.nOutputPlane, net.kW))
   elseif t == 'nn.SpatialConvolutionMM_dsparse' then
      print(('conv_dsparse(in=%d, out=%d, k=%d, s=%d)'):format(net.nInputPlane, net.nOutputPlane, net.kW, net.sW))
   elseif t == 'cudnn.SpatialMaxPooling' then
      print(('max_pool(k=%d, d=%d)'):format(net.kW, net.dW))
   elseif t == 'nn.StereoJoin' then
      print(('StereoJoin(%d)'):format(net.disp_max))
   elseif t == 'nn.Margin2' then
      print(('Margin2(margin=%f, pow=%d)'):format(opt.m, opt.pow))
   elseif t == 'nn.GHCriterion' then
      print(('GHCriterion(m_pos=%f, m_neg=%f, pow=%d)'):format(opt.m_pos, opt.m_neg, opt.pow))
   elseif t == 'nn.Sequential' then
      for i = 1,#net.modules do
         print_net(net.modules[i])
      end
   else
      print(net)
   end
end

function clean_net(net)
   net.output = torch.CudaTensor()
   net.gradInput = nil
   net.weight_v = nil
   net.bias_v = nil
   net.gradWeight = nil
   net.gradBias = nil
   net.iDesc = nil
   net.oDesc = nil
   net.finput = torch.CudaTensor()
   net.fgradInput = torch.CudaTensor()
   net.tmp_in = torch.CudaTensor()
   net.tmp_out = torch.CudaTensor()
   if net.modules then
      for _, module in ipairs(net.modules) do
         clean_net(module)
      end
   end
   return net
end

function save_net(epoch)
   if arch == 'slow' then
      obj = {clean_net(net_te), clean_net(net_te2), opt}
   elseif arch == 'fast' then
      obj = {clean_net(net_te), opt}
   end
   if epoch == 0 then
      fname = ('net/net_%s.t7'):format(cmd_str)
   else
      fname = ('net/net_%s_%d.t7'):format(cmd_str, epoch)
   end
   torch.save(fname, obj, 'ascii')
   return fname
end

function forward_free(net, input)
   local currentOutput = input
   for i=1,#net.modules do
      net.modules[i].oDesc = nil
      local nextOutput = net.modules[i]:updateOutput(currentOutput)
      if currentOutput:storage() ~= nextOutput:storage() then
         currentOutput:storage():resize(1)
         currentOutput:resize(0)
      end
      currentOutput = nextOutput
   end
   net.output = currentOutput
   return currentOutput
end

function fix_border(net, vol, direction)
   local n = (get_window_size(net) - 1) / 2
   for i=1,n do
      vol[{{},{},{},direction * i}]:copy(vol[{{},{},{},direction * (n + 1)}])
   end
end

function stereo_predict(x_batch, id)
   local vols, vol

   if arch == 'fast' then
      forward_free(net_te_val, x_batch:clone())
      vols = torch.CudaTensor(2, disp_max, x_batch:size(3), x_batch:size(4)):fill(0 / 0)
      adcensus.StereoJoin(net_te_val.output[{{1}}], net_te_val.output[{{2}}], vols[{{1}}], vols[{{2}}])
      fix_border(net_te_val, vols[{{1}}], -1)
      fix_border(net_te_val, vols[{{2}}], 1)
      clean_net(net_te_val)
   end

   disp = {}
   local mb_directions = opt.a == 'predict' and {1, -1} or {-1}
   for _, direction in ipairs(dataset == 'mb' and mb_directions or {1, -1}) do

      
      if arch == 'fast' or arch == 'ad' or arch == 'census' then
         vol = vols[{{direction == -1 and 1 or 2}}]
      end

      _, d = torch.min(vol, 2)
      disp[direction == 1 and 1 or 2] = d:cuda():add(-1)
   end

   collectgarbage()

   --disp[2] = adcensus.median2d(disp[2], 5)

   return disp[2]
end

if opt.a == 'train_tr' or opt.a == 'train_all' or opt.a == 'time' then
   function mul32(a,b)
      return {a[1]*b[1]+a[2]*b[4], a[1]*b[2]+a[2]*b[5], a[1]*b[3]+a[2]*b[6]+a[3], a[4]*b[1]+a[5]*b[4], a[4]*b[2]+a[5]*b[5], a[4]*b[3]+a[5]*b[6]+a[6]}
   end

   function make_patch(src, dst, dim3, dim4, scale, phi, trans, hshear, brightness, contrast)
      local m = {1, 0, -dim4, 0, 1, -dim3}
      m = mul32({1, 0, trans[1], 0, 1, trans[2]}, m) -- translate
      m = mul32({scale[1], 0, 0, 0, scale[2], 0}, m) -- scale
      local c = math.cos(phi)
      local s = math.sin(phi)
      m = mul32({c, s, 0, -s, c, 0}, m) -- rotate
      m = mul32({1, hshear, 0, 0, 1, 0}, m) -- shear
      m = mul32({1, 0, (ws - 1) / 2, 0, 1, (ws - 1) / 2}, m)
      m = torch.FloatTensor(m)
      cv.warp_affine(src, dst, m)
      dst:mul(contrast):add(brightness)
   end

   -- subset training dataset
   if opt.subset < 1 then
      function sample(xs, p)
         local perm = torch.randperm(xs:nElement()):long()
         return xs:index(1, perm[{{1, xs:size(1) * p}}])
      end

      local tr_subset
      if dataset == 'kitti' or dataset == 'kitti2015' or dataset == 'modelfish' then
         tr_subset = sample(tr, opt.subset)
      end

      local nnz_tr_output = torch.FloatTensor(nnz_tr:size()):zero()
      local t = adcensus.subset_dataset(tr_subset, nnz_tr, nnz_tr_output);
      nnz_tr = nnz_tr_output[{{1,t}}]
   end
   collectgarbage()

   if opt.a == 'train_all' then
      nnz = torch.cat(nnz_tr, nnz_te, 1)
   elseif opt.a == 'train_tr' or opt.a == 'time' then
      nnz = nnz_tr
   end

   if opt.a ~= 'time' then
      perm = torch.randperm(nnz:size(1))
   end

   local fm = torch.totable(torch.linspace(opt.fm, opt.fm, opt.l1):int())

   -- network for training
   
   if arch == 'fast' then
      net_tr = nn.Sequential()
      for i = 1,#fm do
         net_tr:add(cudnn.SpatialConvolution(i == 1 and n_input_plane or fm[i - 1], fm[i], opt.ks, opt.ks))
         if i < #fm then
            net_tr:add(cudnn.ReLU(true))
         end
      end
      net_tr:add(nn.Normalize2())
      net_tr:add(nn.StereoJoin1())
      net_tr:cuda()

      net_te = net_tr:clone('weight', 'bias')
      net_te.modules[#net_te.modules] = nn.StereoJoin(1):cuda()
      for i = 1,#net_te.modules do
         local m = net_te:get(i)
         if torch.typename(m) == 'cudnn.SpatialConvolution' then
            m.padW = 1
            m.padH = 1
         end
      end

      criterion = nn.Margin2(opt.m, opt.pow):cuda()
   end

   print_net(net_tr)

   params = {}
   grads = {}
   momentums = {}
   for i = 1,net_tr:size() do
      local m = net_tr:get(i)
      if m.weight then
         m.weight_v = torch.CudaTensor(m.weight:size()):zero()
         table.insert(params, m.weight)
         table.insert(grads, m.gradWeight)
         table.insert(momentums, m.weight_v)
      end
      if m.bias then
         m.bias_v = torch.CudaTensor(m.bias:size()):zero()
         table.insert(params, m.bias)
         table.insert(grads, m.gradBias)
         table.insert(momentums, m.bias_v)
      end
   end

   ws = get_window_size(net_tr)
   x_batch_tr = torch.CudaTensor(opt.bs * 2, n_input_plane, ws, ws)
   y_batch_tr = torch.CudaTensor(opt.bs)
   x_batch_tr_ = torch.FloatTensor(x_batch_tr:size())
   y_batch_tr_ = torch.FloatTensor(y_batch_tr:size())

   time = sys.clock()
   for epoch = 1,opt.epochs do
      if opt.a == 'time' then
         break
      end
      if epoch == 15 then
         opt.lr = opt.lr / 2
      end
      if epoch > 15 and epoch % opt.epoch_step == 0 then opt.lr = opt.lr/2 end

      local err_tr = 0
      local err_tr_cnt = 0
      for t = 1,nnz:size(1) - opt.bs/2,opt.bs/2 do
         for i = 1,opt.bs/2 do
            d_pos = torch.uniform(-opt.true1, opt.true1)
            d_neg = torch.uniform(opt.false1, opt.false2)
            if torch.uniform() < 0.5 then
               d_neg = -d_neg
            end

            assert(opt.hscale <= 1 and opt.scale <= 1)
            local s = torch.uniform(opt.scale, 1)
            local scale = {s * torch.uniform(opt.hscale, 1), s}
            if opt.hflip == 1 and torch.uniform() < 0.5 then
               scale[1] = -scale[1]
            end
            if opt.vflip == 1 and torch.uniform() < 0.5 then
               scale[2] = -scale[2]
            end
            local hshear = torch.uniform(-opt.hshear, opt.hshear)
            local trans = {torch.uniform(-opt.trans, opt.trans), torch.uniform(-opt.trans, opt.trans)}
            local phi = torch.uniform(-opt.rotate * math.pi / 180, opt.rotate * math.pi / 180)
            local brightness = torch.uniform(-opt.brightness, opt.brightness)
            assert(opt.contrast >= 1 and opt.d_contrast >= 1)
            local contrast = torch.uniform(1 / opt.contrast, opt.contrast)

            local scale_ = {scale[1] * torch.uniform(opt.d_hscale, 1), scale[2]}
            local hshear_ = hshear + torch.uniform(-opt.d_hshear, opt.d_hshear)
            local trans_ = {trans[1], trans[2] + torch.uniform(-opt.d_vtrans, opt.d_vtrans)}
            local phi_ = phi + torch.uniform(-opt.d_rotate * math.pi / 180, opt.d_rotate * math.pi / 180)
            local brightness_ = brightness + torch.uniform(-opt.d_brightness, opt.d_brightness)
            local contrast_ = contrast * torch.uniform(1 / opt.d_contrast, opt.d_contrast)

            local ind = perm[t + i - 1]
            img = nnz[{ind, 1}]
            dim3 = nnz[{ind, 2}]
            dim4 = nnz[{ind, 3}]
            d = nnz[{ind, 4}]
            if dataset == 'modelfish' then
               x0 = X0[img]
               x1 = X1[img]
            end

            make_patch(x0, x_batch_tr_[i * 4 - 3], dim3, dim4, scale, phi, trans, hshear, brightness, contrast)
            make_patch(x1, x_batch_tr_[i * 4 - 2], dim3, dim4 - d + d_pos, scale_, phi_, trans_, hshear_, brightness_, contrast_)
            make_patch(x0, x_batch_tr_[i * 4 - 1], dim3, dim4, scale, phi, trans, hshear, brightness, contrast)
            make_patch(x1, x_batch_tr_[i * 4 - 0], dim3, dim4 - d + d_neg, scale_, phi_, trans_, hshear_, brightness_, contrast_)

            y_batch_tr_[i * 2 - 1] = 0
            y_batch_tr_[i * 2] = 1
         end

         x_batch_tr:copy(x_batch_tr_)
         y_batch_tr:copy(y_batch_tr_)

         for i = 1,#params do
            grads[i]:zero()
         end

         net_tr:forward(x_batch_tr)

         local err = criterion:forward(net_tr.output, y_batch_tr)
         if err >= 0 and err < 100 then
            err_tr = err_tr + err
            err_tr_cnt = err_tr_cnt + 1
         else
            print(('WARNING! err=%f'):format(err))
         end

         criterion:backward(net_tr.output, y_batch_tr)
         net_tr:backward(x_batch_tr, criterion.gradInput)

         for i = 1,#params do
            momentums[i]:mul(opt.mom):add(-opt.lr, grads[i])
            params[i]:add(momentums[i])
         end

      end



       --Validation
       	 --net_te = net_tr:clone('weight', 'bias')
         val_fname = save_net(epoch)
         net_te_val = torch.load(val_fname, 'ascii')[1]
      	 net_te_val.modules[#net_te_val.modules] = nil

       --Train accuracy
		examples = torch.totable(tr)
      	err_sum = 0
		x_batch = torch.CudaTensor(2, 1, height, width)
		pred_good = torch.CudaTensor()
		pred_bad = torch.CudaTensor()
		for _, i in ipairs(examples) do
		      img_height = metadata[{i,1}]
		      img_width = metadata[{i,2}]
		      id = metadata[{i,3}]
		      x0 = X0[{i,{},{1,img_height},{1,img_width}}]
		      x1 = X1[{i,{},{1,img_height},{1,img_width}}]

		   x_batch:resize(2, 1, img_height, img_width)
		   x_batch[1]:copy(x0)
		   x_batch[2]:copy(x1)

		   collectgarbage()
		   cutorch.synchronize()
		   sys.tic()
		   pred = stereo_predict(x_batch, id)
		   cutorch.synchronize()
		   runtime = sys.toc()


		      assert(not isnan(pred:sum()))

		      actual = dispnoc[{i,{},{1,img_height},{1,img_width}}]:cuda()

		      pred_good:resizeAs(actual)
		      pred_bad:resizeAs(actual)
		      mask = torch.CudaTensor():resizeAs(actual):ne(actual, 0)
		      actual:add(-1, pred):abs()
		      pred_bad:gt(actual, err_at):cmul(mask)
		      pred_good:le(actual, err_at):cmul(mask)
		      local err = pred_bad:sum() / mask:sum()
		      err_sum = err_sum + err
		end
		mean_train_acc = (1.0-err_sum / #examples)
		print('Training accuracy', mean_train_acc,'|',epoch)


       --val accuracy 
      	examples = torch.totable(te)
      	err_sum = 0
		x_batch = torch.CudaTensor(2, 1, height, width)
		pred_good = torch.CudaTensor()
		pred_bad = torch.CudaTensor()
		for _, i in ipairs(examples) do
		      img_height = metadata[{i,1}]
		      img_width = metadata[{i,2}]
		      id = metadata[{i,3}]
		      x0 = X0[{i,{},{1,img_height},{1,img_width}}]
		      x1 = X1[{i,{},{1,img_height},{1,img_width}}]

		   x_batch:resize(2, 1, img_height, img_width)
		   x_batch[1]:copy(x0)
		   x_batch[2]:copy(x1)

		   collectgarbage()
		   cutorch.synchronize()
		   sys.tic()
		   pred = stereo_predict(x_batch, id)
		   cutorch.synchronize()
		   runtime = sys.toc()


		      assert(not isnan(pred:sum()))

		      actual = dispnoc[{i,{},{1,img_height},{1,img_width}}]:cuda()

		      pred_good:resizeAs(actual)
		      pred_bad:resizeAs(actual)
		      mask = torch.CudaTensor():resizeAs(actual):ne(actual, 0)
		      actual:add(-1, pred):abs()
		      pred_bad:gt(actual, err_at):cmul(mask)
		      pred_good:le(actual, err_at):cmul(mask)
		      local err = pred_bad:sum() / mask:sum()
		      err_sum = err_sum + err
		end
		mean_val_acc = (1.0-err_sum / #examples)
		print('Validation accuracy', mean_val_acc,'|',epoch)
		print('Learning rate',opt.lr)


		if testLogger then
        testLogger:add{mean_train_acc, mean_val_acc}
        testLogger:style{'-','-'}
        testLogger:plot()
        
        os.execute('convert -density 200 net/test.log.eps net/test.png')
    	end

     	collectgarbage()
   end
   opt.net_fname = save_net(0)
   collectgarbage()
end

